#pragma once
#include <vector>

const int TAG_VEC_ENCRYPTED_Y = 0;
const int TAG_VEC_ENCRYPTED_GRAD = 1;
const int TAG_VEC_ENCRYPTED_HESS = 2;
const int TAG_SEARCH_RESULTS = 3;
const int TAG_INSTANCE_SPACE = 4;
const int TAG_ENCRYPTED_INSTANCE_SPACE = 5;
const int TAG_BEST_SPLIT_COL_ID = 6;
const int TAG_BEST_SPLIT_THRESHOLD_ID = 7;
const int TAG_BEST_INSTANCE_SPACE = 8;
const int TAG_BEST_PARTY_ID = 9;
const int TAG_DEPTH = 10;
const int TAG_PUBLICKEY = 11;
const int TAG_RECORDID = 12;
const int TAG_ISLEAF = 13;
const int TAG_ISLEFT = 14;
const int TAG_NODE_PARTY_ID = 15;
const int TAG_RECORD_ID = 16;
const int TAG_ENCRYPTED_Y = 17;
