"""Subpackage for evasion attack, which creates a malicious data that the target
machine learning model cannot correctly classify.
"""
from .evasion_attack import Evasion_attack_sklearn  # noqa: F401
