"""Subpackge implementing abstract classes for collaborative learning.
"""
from .api import BaseFedAPI, BaseFLKnowledgeDistillationAPI  # noqa : F401
from .client import BaseClient  # noqa: F401
from .server import BaseServer  # noqa: F401
